#pragma comment(linker, "/version:1.0") // MUST be in the form of major.minor

#include <windows.h>
#include <stdio.h>
#include <tchar.h>
#include "libs\SimpleString.h"

typedef struct {
	BOOL fRecurse;       // recurse directories?
	UINT cSuccessFiles;  // files processed
	UINT cSuccessDirs;   // directories processed
	UINT cErrors;        // errors encountered
} WORKCTX, *PWORKCTX;

VOID WINAPI RemoveFileShortName( PWORKCTX pctx, PCTSTR pszPath, BOOL fIsDirectory );
VOID WINAPI ProcessSearchPath( PWORKCTX pctx, PCTSTR pszSearchPath );
VOID __fastcall PrintError( PCTSTR pszPath, DWORD dwErrorCode );
__forceinline BOOLEAN IsFlag( PCTSTR pszArg );
__forceinline BOOLEAN CheckFlagI( PCTSTR pszArg, TCHAR ch );
__forceinline BOOLEAN IsSpecialDirectoryName( PCTSTR pszPath );
__forceinline BOOLEAN RequestPrivileges( );

UINT __fastcall EntryMain( UINT argc, PTSTR *argv )
{
	UINT iSourceStart = 0;
	BOOLEAN fShowUsage = FALSE;

	// Parse command line parameters

	BOOLEAN fFlagsDone = FALSE;
	UINT i;

	WORKCTX ctx;
	ZeroMemory(&ctx, sizeof(ctx));

	for (i = 1; i < argc; ++i)
	{
		if (fFlagsDone || !IsFlag(argv[i]))
		{
			fFlagsDone = TRUE;

			if (iSourceStart == 0)
				iSourceStart = i;
		}
		else if (!ctx.fRecurse && CheckFlagI(argv[i], TEXT('r')))
			ctx.fRecurse = TRUE;
		else
			fShowUsage = TRUE;
	}

	// Act according to the command line parameters

	if (fShowUsage || iSourceStart == 0)
	{
		static const TCHAR szUsageTemplateGeneral[] = TEXT("  -%c  %s.\n");

		_tprintf(TEXT("Usage: stripshortname [-r] filenames\n\n")
		         TEXT("Options:\n"));
		_tprintf(szUsageTemplateGeneral, TEXT('r'), TEXT("Recursive mode; processes directory trees"));

		return(~1);
	}
	else if (!RequestPrivileges())
	{
		PrintError(TEXT("RequestPrivileges"), GetLastError());
		++ctx.cErrors;
	}
	else
	{
		for (i = iSourceStart; i < argc; ++i)
		{
			ProcessSearchPath(&ctx, argv[i]);
		}

		// Post-mortem
		_tprintf(TEXT("\t%d file(s) and %d folder(s) processed"), ctx.cSuccessFiles, ctx.cSuccessDirs);
		if (ctx.cErrors) _tprintf(TEXT(", with %d error(s)"), ctx.cErrors);
		_tprintf(TEXT(".\n"));
	}

	return(ctx.cErrors);
}

VOID WINAPI RemoveFileShortName( PWORKCTX pctx, PCTSTR pszPath, BOOL fIsDirectory )
{
	BOOLEAN fSuccess = FALSE;

	HANDLE hFile = CreateFile(
		pszPath,
		GENERIC_WRITE | DELETE,
		FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,
		NULL,
		OPEN_EXISTING,
		FILE_FLAG_BACKUP_SEMANTICS,
		NULL
	);

	if (hFile != INVALID_HANDLE_VALUE)
	{
		if (SetFileShortName(hFile, TEXT("")))
			fSuccess = TRUE;

		CloseHandle(hFile);
	}

	if (fSuccess)
	{
		if (fIsDirectory)
			++pctx->cSuccessDirs;
		else
			++pctx->cSuccessFiles;
	}
	else
	{
		++pctx->cErrors;
		PrintError(pszPath, GetLastError());
	}
}

VOID WINAPI ProcessSearchPath( PWORKCTX pctx, PCTSTR pszSearchPath )
{
	HANDLE hFind;
	WIN32_FIND_DATA finddata;

	// Copy the path to a scratchpad buffer and establish the append position
	TCHAR szPath[MAX_PATH << 2];
	PTSTR pszPathAppend;

	lstrcpy(szPath, pszSearchPath);

	if (pszPathAppend = _tcsrchr(szPath, TEXT('\\')))
		++pszPathAppend;
	else
		pszPathAppend = szPath;

	// Special handling is required if the user explicitly specifies "." or ".."
	if (IsSpecialDirectoryName(pszSearchPath))
	{
		RemoveFileShortName(pctx, szPath, TRUE);

		if (pctx->fRecurse)
		{
			lstrcat(pszPathAppend, TEXT("\\*"));
			ProcessSearchPath(pctx, szPath);
		}
	}

	// General case
	else if ((hFind = FindFirstFile(pszSearchPath, &finddata)) != INVALID_HANDLE_VALUE)
	{
		do
		{
			if (!IsSpecialDirectoryName(finddata.cFileName))
			{
				lstrcpy(pszPathAppend, finddata.cFileName);

				RemoveFileShortName(pctx, szPath, finddata.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY);

				if (pctx->fRecurse && finddata.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
				{
					lstrcat(pszPathAppend, TEXT("\\*"));
					ProcessSearchPath(pctx, szPath);
				}
			}

		} while (FindNextFile(hFind, &finddata));

		FindClose(hFind);
	}
}

VOID __fastcall PrintError( PCTSTR pszPath, DWORD dwErrorCode )
{
	_ftprintf(stderr, TEXT("Error: %s"), pszPath);

	if (dwErrorCode)
	{
		PTSTR pszErrorMessage = NULL;

		DWORD cchMessage = FormatMessage(
			FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_FROM_SYSTEM,
			NULL,
			dwErrorCode,
			0,
			(PTSTR)&pszErrorMessage,
			0,
			NULL
		);

		if (cchMessage)
		{
			pszErrorMessage[_tcscspn(pszErrorMessage, TEXT("\r\n"))] = 0;
			_ftprintf(stderr, TEXT(" [%s]"), pszErrorMessage);
		}
		else
		{
			_ftprintf(stderr, TEXT(" [%08X]"), dwErrorCode);
		}

		LocalFree(pszErrorMessage);
	}

	_ftprintf(stderr, TEXT("\n"));
}

__forceinline BOOLEAN IsFlag( PCTSTR pszArg )
{
	return(
		(pszArg[0] | 0x02) == TEXT('/') &&
		(pszArg[1]       ) != 0 &&
		(pszArg[2]       ) == 0
	);
}

__forceinline BOOLEAN CheckFlagI( PCTSTR pszArg, TCHAR ch )
{
	return(
		(pszArg[1] | 0x20) == ch
	);
}

__forceinline BOOLEAN IsSpecialDirectoryName( PCTSTR pszPath )
{
	// TRUE if name is "." or ".."

	#ifdef UNICODE
	return(
		(*((UPDWORD)pszPath) == WCHARS2DWORD(L'.', 0)) ||
		(*((UPDWORD)pszPath) == WCHARS2DWORD(L'.', L'.') && pszPath[2] == 0)
	);
	#else
	return(
		(*((UPWORD)pszPath) == CHARS2WORD('.', 0)) ||
		(*((UPWORD)pszPath) == CHARS2WORD('.', '.') && pszPath[2] == 0)
	);
	#endif
}

__forceinline BOOLEAN RequestPrivileges( )
{
	BOOLEAN fSuccess = FALSE;

	HANDLE hToken;
	TOKEN_PRIVILEGES tkp;

	if (OpenProcessToken(GetCurrentProcess(), TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY, &hToken))
	{
		LookupPrivilegeValue(NULL, SE_RESTORE_NAME, &tkp.Privileges[0].Luid);
		tkp.PrivilegeCount = 1;
		tkp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;

		if (AdjustTokenPrivileges(hToken, FALSE, &tkp, 0, NULL, NULL))
			fSuccess = TRUE;

		CloseHandle(hToken);
	}

	return(fSuccess);
}
